 ;read_dump_bin.pro
;
; This procedure reads in the Binary dump files from the Aerospace Binary Data Extractor.  This procedure uses the Aerospace functaion read_extraction.pro
;
; INPUTS
;
; DUMP_FILE = a string, the name (with full path eg. '/home/tschwart/FP-14/u2001620_EV_C001_M.bin' of the dumpfile to be read in.
;   
; OPTIONAL INPUTS
; 
; MSCAN_NUM = an integer, the maximum scan number in the input file.  Needed to pass to read_extraction.pro.  Default is 32, if there are more this must be specified.
; MSAMP_NUM = an integer, the maximum sample number in the input file.  Needed to pass to read_extraction.pro.  Default is 4095, if there are more this must be specified.
;
; OUTPUTS 
; 
; SCAN =  a long array.  The scan number from the dump file
; HAM = An byte array, the HAM number column from the dump file.
; DET = An integer array, the detector number from the dump file.
; SAMPLE = An integer array, the sample, or frame number from the dump file.
; GAIN = An byte array, the gain state (0, or 1) from the dump file
; DN = An double array, the digital number from the dump file
;
; OPTIONAL OUTPUTS
;
; UAID = a long array the uaid column  from the dump file
; COllECT = an integer array, the collect column from the dump file
; APID = An integer array, the packet ID number from the dump file
; BAND = A string array, the band name column from the dump file.
; DNB_S = An integer array, the DNB_S column from the dump file.  This curently isn't implemented and an array of zeroes will be returned.
; DSECT = An integer array, the sector number from the dump file.
; OFFSET = A long array, the offset from the dump file.
; ABS_SCAN = A long array, same as scan but with the scanoffset added.  If all scans are dumped and the absolute scan number is specified then this array will be the
;	absolute scan number and then scan will be the relative scan number.
;
; REQUIRED FILES
; 
; read_extraction.pro - Aerospace utility to read in extraction file.
;
; OUTPUT Files 
;
; This procedure will read in a dump file produced by the Raytheon Data extractor tool and return each column as an array.  If the option keyword SAVE_FILE is specified
; an IDL save file will be produced with the same arrays the program outputs. The optional outputs will be put into the save file even if not specified.
;
; KNOWN ISSUES
; 
; The keywork DNB_S is not functional.  I don't know where to find this data in the extraction structure.
;
; Version 1 - Tom Schwarting NASA/NICST 2-2009
; Version 2 - Jeff McIntire NASA/NICST 4-2009
;************************************************************************************************************************************************************
;************************************************************************************************************************************************************
;Read_Dump
PRO READ_DUMP_BIN,DUMP_FILE,SCAN,HAM,DET,SAMPLE,GAIN,DN,UAID=UAID,COLLECT=COLLECT,APID=APID,BAND=BAND,DNB_S=DNB_S,$
	DSECT=DSECT,OFFSET=OFFSET,MSCAN_NUM=MSCAN_NUM,MSAMP_NUM=MSAMP_NUM,ABS_SCAN=ABS_SCAN

band_names = ['NOTUSED','DNB','M1','M2','M4','M3','I1','I2','M7','M5','M6','M13','M12','I4','I3','M10','M11','M8', $
              'M9','I5','M16A','M16B','M15','M14','M16AB','DNBMGS','DNBLGS','DNBHGA','DNBHGB','I1OA','I1OB','I1EA', $
              'I1EB','I2OA','I2OB','I2EA','I2EB','I4OA','I4OB','I4EA','I4EB','I3OA','I3OB','I3EA','I3EB','I5OA','I5OB','I5EA','I5EB']

;*****************************************************************************************************************
;CHECK INPUTS
;*****************************************************************************************************************

;Check UAID INPUT  
 HELP,dump_file,OUTPUT=input_check
 input_check=STRSPLIT(input_check,' ',/EXTRACT)
 IF input_check[1] NE 'STRING' THEN PRINT, 'ERROR: Invalid DUMP_FILE!!'

; Check MSCAN_NUM
IF N_ELEMENTS(mscan_num) EQ 0 THEN mscan_num=32

; Check MSAMP_NUM
IF N_ELEMENTS(msamp_num) EQ 0 THEN msamp_num=4095

;*****************************************************************************************************************
;FINISHED CHECKING INPUTS
;*****************************************************************************************************************

;First use Read_Extraction to read in the data
data=READ_EXTRACTION(dump_file,MAXSCANNUMBER=mscan_num,MAXNUMSAMPLES=msamp_sun)

;First see if there is data in the dump file for this band/view
bi=WHERE(data.numscans NE 0)
vi=WHERE(data.numsamples[bi,*] NE 0)

;make output arrays
dn=FLTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
sample=LONARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
scan=LONARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
ham=BYTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
det=INTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
gain=BYTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi]) 

;optional outputs
uaid=LONARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
collect=INTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
apid=INTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
band=STRARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
dnb_s=INTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
dsect=INTARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
offset=LONARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])
abs_scan = LONARR(data.numscans[bi]*data.numdetectors[bi]*data.numsamples[bi,vi])

bi=bi[0]
vi=vi[0]

;dereference pointer to access heap variables
tempdns=*data.dn[bi,vi]
tempgains=*data.gain[bi,vi]

;make the dn array indices varrying fastest to slowest - sample>det>scan
FOR scan_in=0,data.numscans[bi]-1 DO BEGIN
	FOR det_in=0,data.numdetectors[bi]-1 DO BEGIN
		
		index_l=scan_in*data.numdetectors[bi]*data.numsamples[bi,vi]+det_in*data.numsamples[bi,vi]
		index_h=scan_in*data.numdetectors[bi]*data.numsamples[bi,vi]+det_in*data.numsamples[bi,vi]+data.numsamples[bi,vi]-1

		;mandatory outputs	
		dn[index_l:index_h]=tempdns[scan_in,det_in,*]
		scan[index_l:index_h]=LONARR(data.numsamples[bi,vi])+scan_in+1
		ham[index_l:index_h]=BYTARR(data.numsamples[bi,vi])+data.hamforscan[bi,scan_in+1]
		det[index_l:index_h]=INTARR(data.numsamples[bi,vi])+det_in+1
		gain[index_l:index_h]=tempgains[scan_in,det_in,*]
		sample[index_l:index_h]=INDGEN(data.numsamples[bi,vi])+data.firstsample[bi,vi]
		
		;optional outputs
		uaid[index_l:index_h]=LONARR(data.numsamples[bi,vi])+data.uaid
		collect[index_l:index_h]=INTARR(data.numsamples[bi,vi])+data.collect
		apid[index_l:index_h]=INTARR(data.numsamples[bi,vi])+data.apids[bi,vi]
		band[index_l:index_h]=STRARR(data.numsamples[bi,vi])+band_names[bi]
		dnb_s[index_l:index_h]=INTARR(data.numsamples[bi,vi])
		dsect[index_l:index_h]=INTARR(data.numsamples[bi,vi])+vi
		offset[index_l:index_h]=INTARR(data.numsamples[bi,vi])+data.scannumberoffset		
                abs_scan[index_l:index_h] = LONARR(data.numsamples[bi,vi])+data.scanatindex[bi,scan_in]+data.scannumberoffset
			
	ENDFOR ; det_in
ENDFOR ; scan_in

;Garbage Collect
data=0.
HEAP_GC

END ; file
