; Read in extracted data from the binary data format.
;
; Input:
; filename:		The name of the file containing the extracted data
; verbose=verbose:
;				Flag indicating that timing and other diagnostic information
;				should be printed
; maxScanNumber=maxScanNumber:
;				The maximum relative scan number contained in the file. This
;				defaults to 32. If the file contains a scan greater than this
;				number, ingestion will fail
; maxNumSamples=maxNumSamples:
;				The maximum number of samples for any scan/band in the file.
;				This defaults to 4096. If the file contains a sequence of
;				samples greater longer than this value, ingestion will fail.
;
; The extraction file may contain multiple bands from a single UAID/collect, but
; for multiple views. The return variable is an anonymous structure with the
; following fields. Many fields are indexed on band numbers. Band numbers are
; 1-based and reflect the contractor definitions (e.g. M1 is band 2). Even
; though IDL arrays are 0-based, indexes into the band arrays use the actual
; band numbers. Arrays indexed by scan are also 1-based indices. The view
; dimension is indexed by the view structure set by band_info.pro.
;
; read_extraction can handle some flexibility, but if multiple views are in
; the file, it assume that the same scans and detectors are present for the
; multiple views.
;
; uaid: 		The numerical value of the UAID ID.
; uaidName:		The string version of the UAID ID, interpreting the bytes of the
;               little endian numerical values as individual characters. If
;				a "string" UAID ID wasn't specified as part of extraction, this
;				field will likely look like garbage.
; collect:		The numerical value of the collect ID.
; collectName:	The string version of the collect ID, interpreting the bytes of
;               the little endian numerical values as individual characters. If
;				a "string" collect ID wasn't specified as part of extraction,
;				this field will likely look like garbage
; bands:		A band and view indexed array, indicating whether each band is
;				contained within the extracted data (value 1) or not (value 0).
;				If you desire a compacted list of band indices, you can:
;					where(data.bands[*,view] ne 0)
; views:		A view indexed array, indicating whether each view is contained
;				within the extracted data (value 1) or not (value 0). If you
;				desire a compacted list of view indices, you can:
;					where(data.views ne 0)
; apids:		A band and view indexed array of the APID values for each
;				contained band
; scanNumberOffset: The number of the scan at index 0 in scanIndices. This
;				will allow either relative or absolute scan numbers in an input
;				file to be used, and indexed into scanIndices, without wasting
;				a lot of space. Note that the scan specified here isn't
;				neccessarily in the file.
; numScans:		A band indexed array specifying the number of scans contained
;				in the data for each band
; scanIndices:	An array indexed by band and then offset scan (specified scan
;				number - scanNumberOffset), specifying the index
;				into the actual data arrays for the scan dimension for that scan
; scanAtIndex:	An array indexed by band and scan indices, specifying the number
;				of the scan contained at that scan index. This array is the
;				reverse of scanIndices.
; hamForScan:	An array indexed by band and offset scan values, specifying the
;				HAM side for that scan
; hamForScanIndex:
;				An array indexed by band and scan index, specifying the HAM
;				side for that scan
; firstDetector:
;				An array indexed by band indicating the number of the first
;				detector found for that band. The data for that detector is
;				stored at detector index 0. The first detector is detector 1.
; numDetectors:	An array indexed by band indicating the number of detectors
;				in the returned data array. If the actual range of detectors
;				in the extracted data is not contiguous, there will be holes
;				in the data arrays
; firstSample:	An array indexed by band and view indicating the first sample in
;				the returned data array. The data for this sample is stored in
;				at sample index 0. The first sample is sample 0
; numSamples:	An array indexed by band and view indicating the number of
;				samples in the returned data array. If the actual range of
;				samples is not contiguous, there will be holes in the data
;				arrays.
; sampleIndexIncrement:
;				An array indexted by band and view indicating whether the
;				samples contained in the file for that view are from
;				a contiguous range (1) or a non-contiguous range (0). If the
;				input file did not contain this information, the value will
;				be -1
; dn:			An array of pointers to 3D arrays containing the DN values for
;				each band/view. The referenced array is indexed by
;				[scan, detector, sample].
; gain:			An array of pointers to 3D arrays containing the gain values
;				for each band/view. The referenced array is index by
;				[scan, detector, sample].
; gainSumBSD:	An array of pointers to 2D arrays containing the summary gain
;				values across samples for a given scan/detector for Earth view.
;				The main array is indexed by [band]. The referenced arrays are
;				sampled by [scan, detector]. This summary value is the mean gain
;				value of all samples present. It should not be used for the DNB.
; gainSumBS:	An array of pointers to 1D arrays containing the summary gain
;				values for Earth view across samples and detectors for a given
;				scan. The main array is indexed by [band]. The referenced arrays
;				are indexed by [scan]. This summary value is the mean gain value
;				of all samples/detectors for that scan. It should not be used
;				for the DNB.
; gainSumB:		An array of summary gain values, indexed by [band], for the
;				Earth view. Each summary value is the mean of all gain values
;				for that band and should not be used for the DNB.
; gainSum:		The summary gain value for all bands for the Earth view.
;
; Examples:
; For band b, view v, scan sc, detector d, sample sa, in a variable called "data",
; the value is stored in
; (*data.dn[b,v])[data.scanIndices[b, sc - data.scanNumberOffset], d - data.firstDetector[b], sa - data.firstSample[b]]

; Small function to skip a specified number of bytes in a file without
; reading the data. IDL includes a skip function, but it does read all the data
; as it skips, which we don't want to do
pro skipBytes, lun, offset
	point_lun, -lun, current
	point_lun, lun, current+offset
end

; Helper function to open a file that may or may not be compressed, allowing
; either the actual file name, or the filename minus .gz to be specified
pro gzopenr_for_re, unit, file, error=err, actualfile=actualfile, _ref_extra=extra

		; Try to open the file using the given name. Note that
		; while we don't expect this file to be gzip compressed,
		; allow for that case in the event the caller passes in the
		; actual name of the gzipped file (file.gz)
		openr, unit, file, error=err, /compress, _extra=extra
		if (err eq 0) then begin
			actualfile=file
			return
		endif

		; The given file failed, so try a gzipped version of that
		compressedfile = file + '.gz'
		openr, unit, compressedfile, error=err, /compress, _extra=extra
		if (err eq 0) then begin
			actualfile = compressedfile
			return
		endif

		; Neither worked, so return the error
		return
end

; Helper function to convert a little endian integer into a 4-character code.
; The first character of the resulting string is found in the low byte of
; the integer
function num_to_fourcc, num
	; First convert the number to a byte array.
	bytes = bytarr(4) ; a long input will produce at most 4 characters
	for i = 0, 3 do begin
		bytes[i] = num and 255
		num = ishft(num, -8)

		; Clear out any non-printable 7-bit ASCII characters
		if ((bytes[i] lt 32) || (bytes[i] gt 126)) then bytes[i] = 0;
	endfor

	; Applying string to an array of bytes will convert each byte into the
	; array into the character with the ASCII value of that byte
	return, string(bytes)
end

function read_extraction, filename, verbose=verbose, maxScanNumber=maxScanNumber, $
	maxNumSamples=maxNumSamples

	startTime = systime(/seconds)

	; Define a limit to the number of scans. Because scan numbers are
	; 1 based, the actual maximum scan number will be 1 less than this limit.
	if keyword_set(maxScanNumber) then begin
		maxNumScans = maxScanNumber + 1
	endif else begin
		maxNumScans = 33
		maxScanNumber = 32
	endelse

	if ~ keyword_set(maxNumSamples) then maxNumSamples = 4096

	; Bring in extraction data structures.
	band_info, bandInfo=bandInfo, viewInfo=viewInfo

	; Define a couple local constants
	magic1 = '31304542'XUL			; Magic number identifier in the file
	magic1a = '32304542'XUL
	magic2 = '30324542'XUL
	null = ptr_new()				; NULL pointer
	maskRoundToTwo = 'FFFFFFFE'XUL	; Mask for truncating to a multiple of 2
	maxulong = 4294967295UL;

	gzopenr_for_re, extFile, filename, /get_lun, error=err, actualfile=actualfile
	if (err ne 0) then begin
		print, "Can't open file: ", filename
		return, 0
	endif
	
	; Read the magic number from the file and verify that it contains data
	magicFromFile = 0UL
	if not eof(extFile) then readu, extFile, magicFromFile else begin
		print, "File ", actualfile, " is corrupt. First line equal EOF."
		free_lun, extFile
		return, 0
	endelse
	
	; Start back at the beginning in order to keep the looping and code simpler
	point_lun, extFile, 0

	; Make sure the magic number is recognized, and use it to generate
	; the header record. The format of the header needs to match the header
	; format on disk.
	if (magicFromFile eq magic1) || (magicFromFile eq magic1a) then begin
		head = $
		{	magic:0UL, $						; 4-byte The magic number
			uaid:0UL, $							; 4-byte numeric UAID
			collect:0UL, $						; 4-byte numeric collect
			sampleCount:0UL, $					; 4-byte number of samples in group
			firstSample:0L, $ 					; 4-byte first sample number
			scan:0U, $							; 2-byte scan number
			ham:0U, $							; 2-byte HAM side identifier
			apid:0U, $							; 2-byte APID
			band:0U, $							; 2-byte band number
			dnb:0U, $							; 2-byte DNB
			detector:0, $	  					; 2-byte detector number
			viewType:0U, $						; 2-byte view type identifier
			offset:0U $							; 2-byte offset value
		}
		
		; Define a look-up table to convert the coded gain values to actual
		; gain values
		swapGainLut = [0, 2, 1, 3]
		
		; Scans roll over at 2^16
		scanRolloverPoint = 65536ULL
    
		; Each block is padded when the number of samples is odd
		padWhenSamplesEvenOdd = 1
		sampleLengthPreAdd = 1
		sampleLengthPostSubtract = 0
	endif else if magicFromFile eq magic2 then begin
		head = $
		{	magic:0UL, $						; 4-byte The magic number
			uaid:0UL, $							; 4-byte numeric UAID
			scan:0UL, $							; 4-byte scan number
			collect:0U, $						; 2-byte numeric collect
			sampleCount:0U, $					; 2-byte number of samples in group
			firstSample:0, $					; 2-byte first sample number
			sampleIndexIncrement:0, $			; 2-byte sample index increment
			apid:0U, $							; 2-byte APID
			ham:0B, $							; 2-byte HAM side identifier
			band:0B, $							; 2-byte band number
			detector:0B, $						; 2-byte detector number
			viewType:0B $						; 2-byte view type identifier
		}
		
		; In binary extraction mode 2, the gain values are directly stored
		; in the file and don't need swapping
		swapGainLut = [0, 1, 2, 3]
		
		; Scans roll over at 2^32
		scanRolloverPoint = 4294967296ULL;

		; Each block is padded when the number of samaples is even
		padWhenSamplesEvenOdd = 0
		sampleLengthPreAdd = 2
		sampleLengthPostSubtract = 1
	endif else begin
		print, "File ", actualfile, " unrecognized."
		free_lun, extFile
		return, 0
	endelse

	; For several other setup steps, we need the first full header from the file
	point_lun, extFile, 0
	readu, extFile, head


	; Define the data structure into which data from the file will be read
	data = $
	{	version:"$Revision$", $					; Revision of read_extraction
		uaid:0UL, $								; UAID
		uaidName:"", $							; UAID in text form
		collect:0UL, $							; Collect
		collectName:"", $						; Collect in text form
		bands:uintarr(bandInfo.num, viewInfo.num), $			; Place holder for array listing which bands are in the data
		views:uintarr(viewInfo.num), $			; Place holder for array listing which views are in the data
		apids:uintarr(bandInfo.num, viewInfo.num), $	; APID number
		scanNumberOffset:0UL, $				; Number of scan at index 0 in scanIndices
		numScans:lonarr(bandInfo.num), $		; Number of scans
		scanIndices:make_array(bandInfo.num, maxNumScans, /ulong, value=maxulong), $
		$										; Array mapping actual scan number to scan index
		scanAtIndex:make_array(bandInfo.num, maxNumScans, /ulong, value=maxulong), $
		$										; Array mapping dn array indices to scan numbers
		hamForScan:make_array(bandInfo.num, maxNumScans, /byte, value=-1), $
		$										; Array specifying HAM side indicator for each scan
		hamForScanIndex:make_array(bandInfo.num, maxNumScans, /byte, value=-1), $
												; Array specifying HAM side for a scan, but indexed by scan index
		firstDetector:intarr(bandInfo.num), $	; First detector
		numDetectors:intarr(bandInfo.num), $	; Number of detectors
		firstSample:lonarr(bandInfo.num, viewInfo.num), $	; First sample
		numSamples:lonarr(bandInfo.num, viewInfo.num), $	; Number of samples
		sampleIndexIncrement:make_array(bandInfo.num, viewInfo.num, /integer, value=-1), $
												; Flag indicating whether samples are contiguous or not
		dn:ptrarr(bandInfo.num, viewInfo.num), $	; Pointer to the resulting DN values
		gain:ptrarr(bandInfo.num, viewInfo.num), $	; Pointer to the resulting gain value

		gainSumBSD:ptrarr(bandInfo.num), $		; Summary of gain states for each detector
		gainSumBS:ptrarr(bandInfo.num), $		; Summary of gain states for each scan
		gainSumB:fltarr(bandInfo.num), $		; Summary of gain states for each band/view
		gainSum:0.0 $							; Summary of gain states across all earth view bands
	}

	; Populate the static fields in the return data
	data.uaid = head.uaid
	data.uaidName = num_to_fourcc(head.uaid)
	data.collect = head.collect
	data.collectName = num_to_fourcc(head.collect)

	; Create and initialize a series of arrays for cataloging the file. There is
	; one element in each of these arrays for each possible band.
	minDetector = make_array(bandInfo.num, /integer, value=1000)
	maxDetector = make_array(bandInfo.num, /integer, value=-1)
	numDetectors = make_array(bandInfo.num, /integer, value=0)
	minSample = make_array(bandInfo.num, viewInfo.num, /integer, value=32767)
	maxSample = make_array(bandInfo.num, viewInfo.num, /integer, value=-1)
	
	; Set the scanNumberOffset field to be 1 less than the first scan
	; number in the file. Before absolute scan numbers were supported, relative
	; scan 1 would always be at index 1 in data.scanIndices. To not break that,
	; we'll put one less than the first scan number in. One element in those
	; arrays is still wasted, but we don't break any code. Also keep around
	; the actual first scan value
	data.scanNumberOffset = head.scan - 1
	firstScan = ulong64(head.scan)

	; We'll need to create a series of temporary arrays for each differently-sized
	; block of samples in the file. IDL binary file IO can only be done if you
	; have an IDL array that matches the size of the block of data from disk
	; to be read. This pointer array is indexed by sample count. At this point,
	; only allocate the array of pointers. As we discover specific sample counts
	; when the file is cataloged, then allocate the actual sample arrays.
	mergedDN = ptrarr(maxNumSamples + 1)

	; Start back at the beginning in order to keep the looping and code simpler
	point_lun, extFile, 0
	
	while ~ eof(extFile) do begin
		; Read the binary header into our previously defined structure
		readu, extFile, head
		
		; Correct the scan for rollover
		scan = ulong64(head.scan)
		if (scan lt firstScan) then scan += scanRolloverPoint
		relativeScan = ulong(scan - data.scanNumberOffset)

		; Make sure the scan number from the header isn't greater than our
		; previously defined limit. If so, we hadn't allocated enough memory.
		; Throw and error in this case
		if relativeScan gt maxScanNumber then begin
			print, "Scan greater than limit (", strtrim(string(relativeScan), 2), " > ", strtrim(string(maxScanNumber), 2), ") detected!"
			heap_free, data
			heap_free, mergedDN
			free_lun, extFile
			return, 1
		endif

		; Make sure the number of sample's isn't greater than the amount of
		; memory to be allocated. If so, throw an error.
		if (head.firstSample + head.sampleCount) gt maxNumSamples then begin
			print, "Sample sequence greater than limit (", strtrim(string(head.sampleCount), 2), " > ", strtrim(string(maxNumSamples), 2), ") detected!"
			heap_free, data
			heap_free, mergedDN
			free_lun, extFile
			return, 2
		endif

		; If this scan for this band had not previously been encountered (the
		; value for this scan/band combination in data.scanIndices is still -1),
		; then start filling out the values in the return data structure for
		; this scan
		if data.scanIndices[head.band, relativeScan] eq maxulong then begin
			; Set the index for this scan to the next available scan index
			; (equal to the number of scans, given 0-based arrays)
			data.scanIndices[head.band, relativeScan] = data.numScans[head.band]

			; Now that the index has been allocated, store that index in the
			; array that's indexed by scan number
			data.scanAtIndex[head.band, data.numScans[head.band]] = relativeScan

			; Set the HAM side values in the arrays (both the array indexed by
			; scan number and the array indexed by scan index)
			data.hamForScan[head.band, relativeScan] = head.ham
			data.hamForScanIndex[head.band, data.numScans[head.band]] = head.ham

			; Increment the scan count
			data.numScans[head.band]++
		endif

		; If the detector specified in the header is larger or smaller than the
		; detectors previously found for this band, adjust the min/max values
		if head.detector lt minDetector[head.band] then minDetector[head.band] = head.detector
		if head.detector gt maxDetector[head.band] then maxDetector[head.band] = head.detector

		; If the number of the first sample specified in the header is smaller
		; than the previous minimum, update the min.
		if head.firstSample lt minSample[head.band, head.viewType] then minSample[head.band, head.viewType] = head.firstSample

		; If the last sample specified by the header is larger than the previous
		; max, then update the max
		lastSample = head.firstSample + head.sampleCount - 1
		if lastSample gt maxSample[head.band, head.viewType] then maxSample[head.band, head.viewType] = lastSample
		
		; If the file is magic2, it contains an increment value for the sample indices. 
		if (head.magic eq magic2) then begin
			data.sampleIndexIncrement[head.band, head.viewType] = head.sampleIndexIncrement
		endif

		; Store the APID and view values from the header
		data.apids[head.band, head.viewType] = head.apid
		data.bands[head.band, head.viewType] = 1;
		data.views[head.viewType] = 1

		; Allocate a temporary array to store the raw coded sample values from
		; the file for the number of samples specified.
		if mergedDN[head.sampleCount] eq null then $
			mergedDN[head.sampleCount] = ptr_new(uintarr(((head.sampleCount + sampleLengthPreAdd) and maskRoundToTwo) - sampleLengthPostSubtract), /no_copy)

		; Skip over the sample values for now. We need to finish cataloging the
		; file before we can allocate memory to hold them in our return structures.
		; Unfortunately, if the file is compressed, we can't just skip over the bytes using point_lun,
		; as when the final point_lun happens to put the pointer at EOF (zero bytes remaining), point_lun fails
		; As it turns out, the performance difference is not measurable, at least for reasonably sized files
		readu, extFile, *mergedDN[head.sampleCount]
	endwhile

	; Populate the return data structure with the detector and sample info, and
	; then create the actual data arrays for each band
	for b = 0, bandInfo.num - 1 do begin
		data.firstDetector[b] = minDetector[b]
		data.numDetectors[b] = maxDetector[b] - minDetector[b] + 1

		for v = 0, viewInfo.num - 1 do begin
			if maxSample[b,v] gt -1 then begin
				data.firstSample[b,v] = minSample[b,v]
				data.numSamples[b,v] = maxSample[b,v] - minSample[b,v] + 1
			endif else begin
				data.firstSample[b,v] = 0
				data.numSamples[b,v] = 0
			endelse

			if (data.numScans[b] gt 0) && (data.numSamples[b,v] gt 0) then begin
				data.dn[b,v] = ptr_new(fltarr(data.numScans[b], data.numDetectors[b], data.numSamples[b,v]), /no_copy)
				data.gain[b,v] = ptr_new(intarr(data.numScans[b], data.numDetectors[b], data.numSamples[b,v]), /no_copy)
			endif
		endfor

	endfor

	; Put the file pointer back at the beginning and actually read the data
	point_lun, extFile, 0

	while ~ eof(extFile) do begin
		; Re-read the header
		readu, extFile, head

		; Correct the scan for rollover
		scan = ulong64(head.scan)
		if (scan lt firstScan) then scan += scanRolloverPoint
		relativeScan = ulong(scan - data.scanNumberOffset)

		; Using the sampleCount field from the header to point to a previously
		; allocated temporary array, read in the correct number of bytes from
		; the file to match the number of samples specified by the header.
		readu, extFile, *mergedDN[head.sampleCount]

		; Determine the detector index and first and last sample indices
		; for the data in this block
		detectorIndex = head.detector - data.firstDetector[head.band]
		firstSample = head.firstSample - data.firstSample[head.band, head.viewType]
		lastSample = firstSample + head.sampleCount - 1

		; The DN and gain values are coded into a single 2-byte value, so
		; they need to be split apart.
		(*data.gain[head.band, head.viewType])[data.scanIndices[head.band, relativeScan], detectorIndex, firstSample:lastSample] = swapGainLut[ishft((*mergedDN[head.sampleCount])[0:head.sampleCount-1], -14)]

		; If the data for this block is not earth view and not one of the DNB's,
		; the values should be divided by 4 to put them into the same range
		; as the earth view DN values. This can be done as the DN value is
		; extracted from the coded value. However, for BE02 files, the earth
		; view non-DNB data should also be divided by 4. But 
		if ((head.magic eq magic2) || (head.magic eq magic1a)) && (head.viewType eq viewInfo.earth) && (~ bandInfo.isDNB[head.band]) then begin
			(*data.dn[head.band, head.viewType])[data.scanIndices[head.band, relativeScan], detectorIndex, firstSample:lastSample] = ((*mergedDN[head.sampleCount])[0:head.sampleCount-1] and '3FFF'X) / 4.0
		endif else if (head.viewType eq viewInfo.earth) || bandInfo.isDNB[head.band] then begin
			(*data.dn[head.band, head.viewType])[data.scanIndices[head.band, relativeScan], detectorIndex, firstSample:lastSample] = (*mergedDN[head.sampleCount])[0:head.sampleCount-1] and '3FFF'X
		endif else (*data.dn[head.band, head.viewType])[data.scanIndices[head.band, relativeScan], detectorIndex, firstSample:lastSample] = ((*mergedDN[head.sampleCount])[0:head.sampleCount-1] and '3FFF'X) / 4.0
	endwhile

	; Close the file
	free_lun, extFile

	; Summarize the gain info
	for b = 0, bandInfo.num - 1 do begin
		if data.gain[b,viewInfo.earth] eq null then continue

    gainSumBSD = total(*data.gain[b,viewInfo.earth], 3) / float(data.numSamples[b,viewInfo.earth])
		gainSumBS = total(gainSumBSD, 2) / float(data.numDetectors[b])
		data.gainSumB[b] = total(gainSumBS) / float(data.numScans[b])
		data.gainSumBSD[b] = ptr_new(gainSumBSD, /no_copy)
		data.gainSumBS[b] = ptr_new(gainSumBS, /no_copy)
	end

	grandTotal = total(data.bands[*,viewInfo.earth])
	if (grandTotal gt 0) then data.gainSum = total(data.gainSumB[*]) / grandTotal

	; Print out the elapsed time if requested
	if keyword_set(verbose) then begin
		print, 'Ingested: ', actualfile, '(', strtrim(systime(/seconds) - startTime, 2), ')'
	endif

	; Free the memory used by the temporary arrays
	heap_free, mergedDN
	heap_free, bandInfo

	return, data
end
